import _regeneratorRuntime from 'babel-runtime/regenerator';

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

/**
 * An asynchronous WebSocket client.
 * @example
 * // Set up connection.
 * const webSocketClient = new WebSocketClient;
 * // Connect.
 * await webSocketClient.connect("ws://www.example.com/", 9000);
 * // Send is synchronous.
 * webSocketClient.send("Hello!");
 * // Receive is asynchronous.
 * console.log(await webSocketClient.receive());
 * // See if there are any more messages received.
 * if (webSocketClient.dataAvailable !== 0) {
 *     console.log(await webSocketClient.receive());
 * }
 * // Close the connection.
 * await webSocketClient.disconnect();
 */
var WebSocketClient = function () {

    /** @public */


    /** @private */


    /** @private */
    function WebSocketClient() {
        _classCallCheck(this, WebSocketClient);

        this._reset();
    }

    /**
     * Whether a connection is currently open.
     * @returns true if the connection is open.
     */


    /** @private */


    /** @private */


    _createClass(WebSocketClient, [{
        key: 'connect',


        /**
         * Sets up a WebSocket connection to specified url. Resolves when the 
         * connection is established. Can be called again to reconnect to any url.
         */
        value: function () {
            var _ref = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee(url, protocols) {
                return _regeneratorRuntime.wrap(function _callee$(_context) {
                    while (1) {
                        switch (_context.prev = _context.next) {
                            case 0:
                                _context.next = 2;
                                return this.disconnect();

                            case 2:
                                this._reset();

                                this._socket = new WebSocket(url, protocols);
                                this._socket.binaryType = 'arraybuffer';

                                _context.next = 7;
                                return this._setupListenersOnConnect();

                            case 7:
                            case 'end':
                                return _context.stop();
                        }
                    }
                }, _callee, this);
            }));

            function connect(_x, _x2) {
                return _ref.apply(this, arguments);
            }

            return connect;
        }()

        /**
         * Send data through the websocket.
         * Must be connected. See {@link #connected}.
         */

    }, {
        key: 'send',
        value: function send(data) {
            if (!this.connected) {
                throw this._closeEvent || new Error('Not connected.');
            }

            this._socket.send(data);
        }

        /**
         * Asynchronously receive data from the websocket.
         * Resolves immediately if there is buffered, unreceived data.
         * Otherwise, resolves with the next rececived message, 
         * or rejects if disconnected.
         * @returns A promise that resolves with the data received.
         */

    }, {
        key: 'receive',
        value: function () {
            var _ref2 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee2() {
                var _this = this;

                var receivePromise;
                return _regeneratorRuntime.wrap(function _callee2$(_context2) {
                    while (1) {
                        switch (_context2.prev = _context2.next) {
                            case 0:
                                if (!(this._receiveDataQueue.length !== 0)) {
                                    _context2.next = 2;
                                    break;
                                }

                                return _context2.abrupt('return', this._receiveDataQueue.shift());

                            case 2:
                                if (this.connected) {
                                    _context2.next = 4;
                                    break;
                                }

                                throw this._closeEvent || new Error('Not connected.');

                            case 4:
                                receivePromise = new Promise(function (resolve, reject) {
                                    _this._receiveCallbacksQueue.push({ resolve: resolve, reject: reject });
                                });
                                _context2.next = 7;
                                return receivePromise;

                            case 7:
                                return _context2.abrupt('return', _context2.sent);

                            case 8:
                            case 'end':
                                return _context2.stop();
                        }
                    }
                }, _callee2, this);
            }));

            function receive() {
                return _ref2.apply(this, arguments);
            }

            return receive;
        }()

        /**
         * Initiates the close handshake if there is an active connection.
         * Returns a promise that will never reject.
         * The promise resolves once the WebSocket connection is closed.
         */

    }, {
        key: 'disconnect',
        value: function () {
            var _ref3 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee3(code, reason) {
                var _this2 = this;

                return _regeneratorRuntime.wrap(function _callee3$(_context3) {
                    while (1) {
                        switch (_context3.prev = _context3.next) {
                            case 0:
                                if (this.connected) {
                                    _context3.next = 2;
                                    break;
                                }

                                return _context3.abrupt('return', this._closeEvent);

                            case 2:
                                _context3.next = 4;
                                return new Promise(function (resolve, reject) {
                                    // It's okay to call resolve/reject multiple times in a promise.
                                    var callbacks = {
                                        resolve: function resolve(dummy) {
                                            // Make sure this object always stays in the queue
                                            // until callbacks.reject() (which is resolve) is called.
                                            _this2._receiveCallbacksQueue.push(callbacks);
                                        },

                                        reject: resolve
                                    };

                                    _this2._receiveCallbacksQueue.push(callbacks);
                                    // After this, we will imminently get a close event.
                                    // Therefore, this promise will resolve.
                                    _this2._socket.close(code, reason);
                                });

                            case 4:
                                return _context3.abrupt('return', _context3.sent);

                            case 5:
                            case 'end':
                                return _context3.stop();
                        }
                    }
                }, _callee3, this);
            }));

            function disconnect(_x3, _x4) {
                return _ref3.apply(this, arguments);
            }

            return disconnect;
        }()

        /**
         * Sets up the event listeners, which do the bulk of the work.
         * @private
         */

    }, {
        key: '_setupListenersOnConnect',
        value: function () {
            var _ref4 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee4() {
                var _this3 = this;

                var socket;
                return _regeneratorRuntime.wrap(function _callee4$(_context4) {
                    while (1) {
                        switch (_context4.prev = _context4.next) {
                            case 0:
                                socket = this._socket;
                                _context4.next = 3;
                                return new Promise(function (resolve, reject) {

                                    var handleMessage = function handleMessage(event) {
                                        var messageEvent = event;
                                        // The cast was necessary because Flow's libdef's don't contain
                                        // a MessageEventListener definition.

                                        if (_this3._receiveCallbacksQueue.length !== 0) {
                                            _this3._receiveCallbacksQueue.shift().resolve(messageEvent.data);
                                            return;
                                        }

                                        _this3._receiveDataQueue.push(messageEvent.data);
                                    };

                                    var handleOpen = function handleOpen(event) {
                                        socket.addEventListener('message', handleMessage);
                                        socket.addEventListener('close', function (event) {
                                            _this3._closeEvent = event;

                                            // Whenever a close event fires, the socket is effectively dead.
                                            // It's impossible for more messages to arrive.
                                            // If there are any promises waiting for messages, reject them.
                                            while (_this3._receiveCallbacksQueue.length !== 0) {
                                                _this3._receiveCallbacksQueue.shift().reject(_this3._closeEvent);
                                            }
                                        });
                                        resolve();
                                    };

                                    socket.addEventListener('error', reject);
                                    socket.addEventListener('open', handleOpen);
                                });

                            case 3:
                            case 'end':
                                return _context4.stop();
                        }
                    }
                }, _callee4, this);
            }));

            function _setupListenersOnConnect() {
                return _ref4.apply(this, arguments);
            }

            return _setupListenersOnConnect;
        }()

        /**
         * @private
         */

    }, {
        key: '_reset',
        value: function _reset() {
            this._receiveDataQueue = [];
            this._receiveCallbacksQueue = [];
            this._closeEvent = null;
        }
    }, {
        key: 'connected',
        get: function get() {
            // Checking != null also checks against undefined.
            return this._socket != null && this._socket.readyState === WebSocket.OPEN;
        }

        /**
         * The number of messages available to receive.
         * @returns The number of queued messages that can be retrieved with {@link #receive}
         */

    }, {
        key: 'dataAvailable',
        get: function get() {
            return this._receiveDataQueue.length;
        }
    }]);

    return WebSocketClient;
}();

export default WebSocketClient;